#!/usr/bin/env python3

"""
Met en forme les objets de facon a les rendre transportable,
de facon securisee et optimise.
"""

__all__ = [ "serialize", "encrypt", "compress",
            "deserialize", "decrypt", "decompress",
            "tools", "constants"]
